#!perl -w
# PovRay 3.5 Color Map Editor V1.0, Kees Moerman, 1 oct 2002
#
# This program creates an interactive PovRay color map editor
#
# This program will help you create and alter colors and
# color maps in Povray files. Use this program to create new
# maps and lists, copy them into your Povray scenes, and
# use this program to alter them later on.
# Supports Filter and Transparancy settings, color maps with 8
# colors.
#
  my $help_text = <<HELP;		# inline help text placed here
# $0, C.M. Moerman (c) October 2002
#
# Syntax: perl colormap.pl
#
HELP
#
# Kees Moerman, Oct 4, 2002
#
# Change list:
#
# To do:
#   add option to add file name on command line, or drag and drop
#

use Tk;				                    # use Tk GUI routines
use Tk::Dialog;			                # esp. Dialog box for display help
use Tk::ROText;			                # and Read-only text box for output panes
use Tk::Balloon;                        # and balloon help
use Tk::ColorEditor;

use strict;

my $top = MainWindow->new();
   $top->title("Color Map Editor, Kees Moerman 2002");
my $ball = $top->Balloon();             # balloon help

my $input;			                    # forward ref. to input handle
my $output;			                    # forward ref. to output handle
my $help_dialog;		                # forward ref. to help_dialog handle

my $i;

my @transparancies;                     # transparancies per color
my @filters;

my ($sc_trans, $sc_filter);             # a few forwards

my $filename = "";                      # globals for file storage
my $file_head = "";
my $file_tail = "";

# ---------------- menu buttons ---------------

my $but_frame = $top->Frame();
   $but_frame->pack(-side => 'top', -anchor => "nw", -expand => 0, -fill => 'x', -padx => 0);


sub button {
    my ($this, $text, $help, $command) = @_;
    my $button = $this->Button(-text => $text, -command => $command);
    my $hotkey = "<Alt-" . lc(substr($text, 0, 1)) . ">";
    
    $top->bind($hotkey, $command );
    $button->configure(-underline => '0');  # Alt-firstchar as hotkey
    $button->pack(-side => 'left');
    $ball->attach($button, -balloonmsg => $help);
    
    return $button;
}

button($but_frame, "Load",   "Load (named) color from file",         \&load_file);
button($but_frame, "Insert", "Insert lines back to loaded file",     \&insert_file);
button($but_frame, "Save",   "Save lines to file",                   \&save_file);
button($but_frame, "GetCode",  "Get color map Povray code",          \&print_colors);
button($but_frame, "Run",    "Execute selected/all",                 \&execline);
button($but_frame, "Quit",   "Stop program",                         sub { $top->destroy; } );


$i= $but_frame->Label(-text => "  Name:")->pack(-side => 'left');
    $ball->attach($i, -balloonmsg => "colormap/list name");
my $entry_name = $but_frame->Entry(-width => 10);
    $ball->attach($entry_name, -balloonmsg => "colormap/list name");
    $entry_name->pack(-side => 'left');


my $colorflag = 0;
$i= $but_frame->Label(-text => "  colorlist?")->pack(-side => 'left');
    $ball->attach($i, -balloonmsg => "Off: color map, On: color list");
my $but_flag = $but_frame->Checkbutton(-variable => \$colorflag);
    $ball->attach($but_flag, -balloonmsg => "Off: color map, On: color list");
    $but_flag->pack(-side => 'left');


my $b_helpprog =
    button($but_frame, "Help", "Display small help text",        sub { $help_dialog->Show(); } );
    $b_helpprog->pack(-side => 'right');


# ---------------- color buttons ---------------

my $cref = $top->ColorEditor();         # start the color editor
   $cref->Show();                       # and pop up

my $col_frame = $top->Frame();          # new frame for color button bar
   $col_frame->pack(-side => 'top', -anchor => "nw", -expand => 0, -fill => 'x', -padx => 0);

my $nr_cols = 8;                        # 8 colors
my @col_ids;                            # ids for color button widgets

foreach $i (0..$nr_cols-1)              # create buttons
{
    $col_ids[$i] = button($col_frame, "$i", "$i", [ \&set_col, $i] );
    $col_ids[$i]->configure(-width => 11, -height => 5, -background => "#C000C000C000");
    $ball->attach($col_ids[$i], -balloonmsg => "Select color $i, double to copy");
    $col_ids[$i]->bind('<Double-Button-1>', [ \&set_double, $i] );
}

my $active_col = 0;                     # currently active color

sub set_col                             # ( id ) --> () button action
{
    my $thiscol = shift;                # activate new color button
    $cref->configure(-widgets => [ $col_ids[$thiscol] ]);
    $col_ids[$active_col]->configure(-relief => 'raised'); # disable previous
    $active_col = $thiscol;
    $col_ids[$active_col]->configure(-relief => 'sunken');
    $sc_filter->set($filters[$active_col]*100);
    $sc_trans->set($transparancies[$active_col]*100);
    &update_bar();                      # update color bars area
}

sub set_double                          # double click button action
{
    my ($widget, $thiscol) = @_;        # activate new color button
    set_col($thiscol);
    $cref->configure(-color => getcolor($thiscol)); # copy color to editor
}

foreach $i (0..$nr_cols)                # initialise transparancy/filter values
{
    $transparancies[$i] = 0; $filters[$i] = 0;
}


# ---------------- color bars and dividers ---------------

my @points;                             # color band division points 0..1
my @pointbars;                          # ids for point bar line canvas items

foreach $i (0..$nr_cols)
{   $points[$i] = $i/($nr_cols-1); }    # pos along axis evenly over (0..1)

my $nr_bars = 128; my $barwidth = 5;
my @bars;                               # ids for color bar canvas items

my $canvas = $top->Canvas(              # create canvas for color bars
        qw/width 50 -height 65 -bd 0 -highlightthickness 0/);
    $canvas->pack(-side => 'top', -anchor => "nw", -expand => 0, -fill => 'x', -padx => 0);


foreach $i (0..$nr_bars)                # create the color bar canvas items
{   $bars[$i] = $canvas->createLine(
        $i*$barwidth, 0, $i*$barwidth, 50, -fill => 'black', -width => 6);
}

foreach $i (0..$nr_cols-1)              # create the division point line canvas items
{   $pointbars[$i] = $canvas->createLine(
        $i*$barwidth, 0, $i*$barwidth, 45, -fill => 'black', -width => 1);

}
                                        # and the current color line marker
$canvas->create(qw/line 0 0 1 1 2 2 3 3 0 0 -fill Red -width 3 -tags active/);


# ---------------- sliders frame ---------------

my $slider_frame = $top->Frame();       # new frame for color button bar
   $slider_frame->pack(-side => 'top', -anchor => "nw", -expand => 0, -fill => 'x', -padx => 0);

my $sc_map = $slider_frame->Scale(qw/-orient horizontal -length 200 -from 0
        -to 1000 -tickinterval 200 -command/ => [\&map_scale, $canvas]);
   $sc_map->set(200);
   $sc_map->pack(qw/-side left/);
   $ball->attach($sc_map, -balloonmsg => "Colormap Division Slider");

sub map_scale                           # scale for shifting the color map
{                                       # division points
    my ($w, $width);
    ($w, $width) = @_;

    $points[$active_col] = $width/1000;
    my $x = $width * ($nr_bars * $barwidth) / 1000;
    $w->coords('active', $x, 20, $x, 55, $x-5, 60, $x+5, 60, $x, 55);
}


# transparancy scale
   $sc_trans = $slider_frame->Scale(qw/-orient horizontal -length 200 -from 0
        -to 100 -tickinterval 20 -command/ => [\&trans_scale, $canvas]);
   $sc_trans->set(0);
   $sc_trans->pack(qw/-side right/);    # and transparency most right
   $ball->attach($sc_trans, -balloonmsg => "Transparancy");

sub trans_scale                         # called when dragging transparency slider
{
    my ($w, $width) = @_;
    $transparancies[$active_col] = $width/100;
}

# filter scale
   $sc_filter = $slider_frame->Scale(qw/-orient horizontal -length 200 -from 0
        -to 100 -tickinterval 20 -command/ => [\&filter_scale, $canvas]);
   $sc_filter->set(0);
   $sc_filter->pack(qw/-side right/);   # the filter slide in the middle
   $ball->attach($sc_filter, -balloonmsg => "Filter value");

sub filter_scale                        # called when dragging filter slider
{
    my ($w, $width) = @_;
    $filters[$active_col] = $width/100;
}

# ---------------- help box etc. -------------------------------

$help_dialog = $top->Dialog(-title => "PovRay 3.5 Color Editor",
           -default_button => "OK", -wraplength => '5i', -text => <<HELP);
Simple PovRay 3.5 Color_map/list editor by Kees Moerman (c) October 2002.
Make colors using the Color Editor widget, save into a new file or copy and paste (using 'GetCode') in your Povray scene.
Later on, reload them into this program, modify and insert back into your scene.
You can give names to the color maps and lists to allow handling of multiple sets in one file.
Check 'colorlist' to save/insert as list, otherwise save will be as color_map.
Single-click a color button to select and modify, double-click to copy color to Color Editor window.
HELP

my $fail_dialog = $top->Dialog(-title => "PovRay 3.5 Color Editor Error",
           -default_button => "OK", -wraplength => '4i',
           -text => "Operation Load or Save did not succeed (no filename selected)");


# --------------------------- output pane --------------------------

$output = $top->Scrolled('ROText',      # Read-only output window (copy allowed)
		-wrap => "none",
		-scrollbars => "osre",
		-width => 50, -height => 12); # scrolled text box
$output->pack(-expand => 1, -fill => 'both');
    $ball->attach($output, -balloonmsg => "Output area for GetCode to copy/paste to other programs");

$output->insert('end', "Output area for status and GetCode command\n");

# redirection combined with tie to Tk text widget:
tie(*STDOUT, 'Tk::Text', $output);


# ---------------- text veld -------------------------------

$input = $top->Scrolled('Text',         # this is the main window
	            -wrap => "none",
	            -scrollbars => "se",
	            -width => 80, -height => 3);
$input->pack(-expand => 1, -fill => 'both');
$input->focus();                        # gets focus on startup
    $ball->attach($input, -balloonmsg => "Input area for additional commands (non-documented feature)");


# ----------------- file load/save functions ------------------------

my $filetypes = [ ['Povray inc files', ['.inc']],
                  ['Povray pov files', ['.pov']],
                  ['All files', ['*']]
                ];

sub getloadname                         # get file name for load, return name
{
    if((length $filename) && ($filename =~ m#(.*)/(.*)# )) # path set before? re-use
    {
        $filename = $top->getOpenFile(-title => "Select file to open",
                            -initialdir => $1, -initialfile => $2,
                            -filetypes => $filetypes, -defaultextension => ".inc");
    }
    else
    {
        $filename = $top->getOpenFile(-title => "Select file to open",
                            -filetypes => $filetypes, -defaultextension => ".inc");
    }
    
    $fail_dialog->configure(-text => "No file name specified or can not read file '$filename'");
    $fail_dialog->Show() if !$filename || ! -r $filename;  # should be readable
    return "" if !$filename || ! -r $filename;  # should be readable

    if (!open(FILE, $filename))         # next read file line by line
    {
        $filename = "";
        $fail_dialog->configure(-text => "Can not read file '$filename' due to $!");
        $fail_dialog->Show();           # should be readable
        return "";
    }
    
    return $filename;
}

my $name = "";                          # below patterns for map/list recogn.
my $map_pattern  = "color_map\\s*{\\s*//\\s*from\\s*colormap.pl\\s*";
my $list_pattern = "//\\s*color\\s*list\\s*from\\s*colormap.pl\\s*";

sub load_colormap                       # inner loop for color map format
{   my $line = shift;
    my @colors;

    $colorflag = 0;
    if (($name eq "") && ($line =~ m#$map_pattern(.*?)\s*$#))
    { $name = $1; $entry_name->insert(0, $name); }

    $i = 0;

    while($line = <FILE>)               # get lines from color map statement
    {
        last if ($line !~ /color\s*rgb/);
        
# line format:    [ %7.4f color rgbft <%8.5f %8.5f %8.5f %7.4f %7.4f> ] // %d\n",
        
        if($line !~ /\s*\[\s*([\.0-9]+)\s*color\s*rgb..\s*<\s*([.0-9]+)\s*([.0-9]+)\s*([.0-9]+)\s*([.0-9]+)\s*([.0-9]+)\s*>\s*\]/)
        {
            $file_head = "";
            $fail_dialog->configure(-text => "Color format not recognised in '$line'");
            close FILE;
            return $fail_dialog->Show();
        }
        
        ($points[$i], $colors[0], $colors[1], $colors[2], $filters[$i], $transparancies[$i]) =
            ($1, $2, $3, $4, $5, $6);
#       printf "read:   [ %7.4f color rgbtf <%8.5f %8.5f %8.5f %7.4f %7.4f> ]\n",
#           $points[$i], $colors[0], $colors[1], $colors[2], $filters[$i], $transparancies[$i];
        my $hexcolor = sprintf("#%02x00%02x00%02x00",
                int($colors[0]*255), int($colors[1]*255), int($colors[2]*255));
            $col_ids[$i]->configure(-background => $hexcolor);
        $i++;
    }
    return $line;
}

sub load_colorlist                      # inner loop for color list format
{   my $line = shift;
    my @colors;

    $colorflag = 0;
    
    if (($name eq "") && ($line =~ m#$list_pattern(.*?)\s*$#))
    { $name = $1; $entry_name->insert(0, $name); }
    $colorflag = 1 if($line =~ /color\s*list/ );
    $colorflag = 0 if($line =~ /color_map/ );

    $i = 0;

    while($line = <FILE>)           # get lines from color map statement
    {
        last if ($line !~ /color\s*rgb/);
        
# line format:    [ %7.4f color rgbft <%8.5f %8.5f %8.5f %7.4f %7.4f> ] // %d\n",
        
        if($line !~ /color\s*rgb..\s*<\s*([.0-9]+)\s*([.0-9]+)\s*([.0-9]+)\s*([.0-9]+)\s*([.0-9]+)\s*>\s*;/)
        {
            $file_head = "";
            $fail_dialog->configure(-text => "Color format not recognised in '$line'");
            close FILE;
            return $fail_dialog->Show();
        }
        
        ($colors[0], $colors[1], $colors[2], $filters[$i], $transparancies[$i]) =
            ($1, $2, $3, $4, $5, $6);
#       printf "read:   [ %7.4f color rgbtf <%8.5f %8.5f %8.5f %7.4f %7.4f> ]\n",
#           $points[$i], $colors[0], $colors[1], $colors[2], $filters[$i], $transparancies[$i];
        my $hexcolor = sprintf("#%02x00%02x00%02x00",
                int($colors[0]*255), int($colors[1]*255), int($colors[2]*255));
            $col_ids[$i]->configure(-background => $hexcolor);
        $i++;
    }
    return $line;
}

sub load_file {				            # read file into input pane
    my $line;
 
    return if getloadname() eq "";
    print "File read: $filename\n";

    $file_head = ""; $file_tail = "";   # needed for insert

    $name = $entry_name->get();         # get color (nap) name

    while($line = <FILE>)               # read up to color map statement from coloredit
    {   last if ($line =~ m#$map_pattern$name#);
        last if ($line =~ m#$list_pattern$name#);
        print "Map skipped: '$1'; "     # just to give indication of other stored colors
            if ($line =~ m#$map_pattern(.*?)\s*$#);
        print "Color skipped: '$1'; "   # just to give indication of other stored colors
            if ($line =~ m#$list_pattern(.*?)\s*$#);
        $file_head .= $line;
    }    
    
    if($line !~ /color_map/ && $line !~ /color\s*list/ )
    {
        $filename = ""; $file_head = ""; # didn't find a thing...
        if($name)
        {   $fail_dialog->configure(-text => "Specified name $name not found in file"); }
        else
        {   $fail_dialog->configure(-text => "Incorrect file format: color map or list not found"); }
        close FILE;
        return $fail_dialog->Show();    # should contain color keyword
    }

    if($line =~ /color_map/) { $line = load_colormap($line); }
    else { $line = load_colorlist($line); } # go process inner loop
    
    if( ($line !~ /}/) || ($i != 8))    # should have read 8 lines and ended with '}'
    {
        $file_head = "";
        $fail_dialog->configure(-text => "Incorrect format: end of color_map or list not found (i = $i) in '$line'");
        close FILE;
        return $fail_dialog->Show();
    }

    while (<FILE>)                      # read tail for later insert
    {
        $file_tail .= $_;
        print "Map not reached: '$1'; " # just to give indication of other stored maps
            if (m#$map_pattern(.*?)\s*$#);
        print "Color not reached: '$1'; "
            if (m#$list_pattern(.*?)\s*$#);
    }
    close(FILE);
    
#   printf "File head: %d, file tail: %d\n", length $file_head, length $file_tail;
}

sub getsavename                         # get file name for save or insert
{
    if((length $filename) && ($filename =~ m#(.*)/(.*)# )) # split path
    {
        $filename = $top->getSaveFile(-title => "Select file to save to",
                            -initialdir => $1, -initialfile => $2,
                            -filetypes => $filetypes, -defaultextension => ".inc");
    }
    else                                # path not known
    {
        $filename = $top->getSaveFile(-title => "Select file to save to",
                            -filetypes => $filetypes, -defaultextension => ".inc");
    }
    $fail_dialog->configure(-text => "No file name specified for Insert");
    $fail_dialog->Show() if !$filename;
    return $filename;
}

sub save_file {				            # save input pane to file
    my $result;

    if(length $file_head > 5)           # was part from larger file
    {                                   # so user might have wanted to insert instead..
        $file_head = "";
        $fail_dialog->configure(-text => "Warning, you might have meant 'Insert'...");
        $fail_dialog->Show();
    }

    $fail_dialog->configure(-text => "No file name specified for Save");
    return if(! &getsavename());

    if(open(OUTPUT, ">$filename"))
    {   print OUTPUT &make_code();      # save results to file in one go
        close(OUTPUT);
    }
}

sub insert_file {			            # save input pane to file
    my $result;

    if(length $file_head < 5)           # was not from a larger file
    {                                   # should have been 'save'
        $file_head = "";
        $fail_dialog->configure(-text => "Not read from larger file, please use 'save' instead");
        return $fail_dialog->Show();
    }
    $fail_dialog->configure(-text => "No file name specified for Load");
    return if(! &getsavename());

    if(open(OUTPUT, ">$filename"))
    {
        print OUTPUT $file_head;
        print OUTPUT &make_code();      # save results to file in one go
        print OUTPUT $file_tail;
        close(OUTPUT);
    }
}

sub execline {				            # execute selected area or all via ^R or Run
	$top->Busy;
	my $tags = $input->tagNextrange('sel', '1.0', 'end');
	$_ = $input->get('sel.first', 'sel.last') if defined $tags;
	$_ = $input->get('1.0', 'end') if !defined $tags;   # all if none selected

	s/^\?/print /;
	my @result = eval;			        # evaluate entered expression

	print STDERR $@ if ($@);		    # Oops, error

	print "\neval returned: " if !/^print/;
	print @result if !/^print/;	    	# only print if statement didn't
	$output->see('end');		        # scrol display to end

	$top->Unbusy;
}	



# -------------------------- color functions -------------------------------

sub getcolor {                          # ( buttonnumber ) -> ( #rrrrggggbbbb )
    my $index = shift;                  # retrieve color from button
#   print " $index ";
    if (($index > $nr_cols-1) || ($index < 0)) { $index = $nr_cols-1; }
    return $col_ids[$index]->cget(-background);
}

my $RED = 0; my $GREEN = 1; my $BLUE = 2;

no strict;

sub get_comp                            # ( color, index ) -> ( component )
{   my ($color, $index) = @_;           # index == red, green, blue
    print "Not valid color '$color' ($index)" if ! ($color =~ /#[0-9a-fA-F]{12}/);
    print "Invalid index '$index'" if ! (($index >=0) && ($index <= 2));
    $color =~ /#(..)..(..)..(..)../;
    return hex(${$index+1}) / 255;      # does not run with 'use strict'...
}

use strict;

sub make_code                           # create Povray code for color_map or list
{   my ($i, $j, $color, @colors, $result, $name);

    $name = $entry_name->get();         # get name from entry box
    $name = "MyColor" if $name eq "";
    $result = "    color_map { // from colormap.pl $name \n";
    $result = "    // color list from colormap.pl $name \n" if $colorflag;
    
    foreach $i (0..$nr_cols-1)
    {
        $color = getcolor($i);
        $color =~ /#(..)..(..)..(..)../;
        $colors[0] = hex($1)/255; $colors[1] = hex($2)/255; $colors[2] = hex($3)/255;
        
        if($colorflag)
        {
          $result .= sprintf "#declare ${name}_$i = color rgbft <%7.4f %7.4f %7.4f %6.3f %6.3f>;\n",
            @colors, $filters[$i], $transparancies[$i];
        }
        else
        {
          $result .= sprintf "        [ %6.3f color rgbft <%7.4f %7.4f %7.4f %6.3f %6.3f> ] // %d\n",
            $points[$i], @colors, $filters[$i], $transparancies[$i], $i;
        }
    }
    $result .= "    }\n" if !$colorflag;
    $result .= "  //  }\n" if $colorflag;
    return $result;
}

sub print_colors { print(make_code()); }

sub min { if( $_[0] > $_[1] ) { return $_[1]; } return $_[0]; }


sub update_bar {
    my ($color1, $color2, $color, $red, $green, $blue, $i, $prevcol, $j2, $j1, $x, $col);

    $points[0] = 0;                     # boundaries are fixed at 0 and 1
    $points[$nr_cols-1] = 1;
    $sc_map->set($points[$active_col] * 1000);   # update slider to curtrent color

    foreach $i (1..$nr_cols-2)          # move adjacent points if needed
    {   if($points[$i] >= $points[$i+1]) { $points[$i+1] = $points[$i] + 0.001; }
    }
    $points[$nr_cols-1] = 1;
    foreach $j1 (1..$nr_cols-2)
    {   $i = $nr_cols-1-$j1;
        if($points[$i] >= $points[$i+1]) { $points[$i] = $points[$i+1] - 0.001; }
    }

    foreach $i (0..$nr_cols-1)          # move slider bars
    {
        $canvas->coords($pointbars[$i],
            $points[$i] * $barwidth * $nr_bars, 5, $points[$i] * $barwidth * $nr_bars,45);
    }
    

    $prevcol = -10;
    foreach $i (0..$nr_bars)            # go update all individual color bars
    {
        $x = $i/$nr_bars;               # pos runs from 0 up to including 1
        $col = $nr_cols-1;              # init at last color
        foreach $j1 (0..$nr_cols-1) { $col = $j1 if $x >= $points[$j1]; } # get first color

        if($col != $prevcol)             # new color? reload
        {   $color1 = getcolor($col); $color2 = getcolor($col+1); $prevcol = $col;
#           print "##$col##";
        }

        $j1 = ($x - $points[$col]) / ($points[$col+1] - $points[$col]);
        $j2 = 1-$j1;

#       print "$i($x) color=$col ratio = $j2/$j1 '$color1' '$color2'\n";
        
        $red    = (get_comp($color1, $RED)*$j2   + get_comp($color2, $RED)*$j1);
        $green  = (get_comp($color1, $GREEN)*$j2 + get_comp($color2, $GREEN)*$j1);
        $blue   = (get_comp($color1, $BLUE)*$j2  + get_comp($color2, $BLUE)*$j1);
        
        $color = sprintf("#%02x00%02x00%02x00", int($red*255), int($green*255), int($blue*255));
        $canvas->itemconfigure($bars[$i], -fill => $color);
    }
}

set_col(0);                             # default active color

# ------------------- main program ---------------------------

$top->repeat(1000, \&update_bar);       # update bar every second

no strict;                              # not needed in command input mode

MainLoop();				                # start 'main program'

# end of Perl script
